
import collections, cbor2
from .types import *

__all__ = (
    'encoder_default',
    'convert_frozendicts_to_maps')


def encoder_default(encoder, value):
    if isinstance(value, Map):
        return encoder.encode_map(value)
    raise cbor2.types.CBOREncodeTypeError('cannot serialize type ' + type(value).__name__)

def convert_frozendicts_to_maps(value):
    value, _ = _convert_frozendicts_to_maps(value)
    return value

def _convert_frozendicts_to_maps(value):
    if isinstance(value, list):
        converted = False
        list_ = []
        for value_ in value:
            value_, value_converted = _convert_frozendicts_to_maps(value_)
            if value_converted:
                converted = True
            list_.append(value_)
        if converted:
            return list_, True

    elif isinstance(value, tuple):
        converted = False
        list_ = []
        for value_ in value:
            value_, value_converted = _convert_frozendicts_to_maps(value_)
            if value_converted:
                converted = True
            list_.append(value_)
        if converted:
            return tuple(list_), True

    elif isinstance(value, collections.abc.Mapping):
        map_ = Map()
        converted = isinstance(value, cbor2.types.FrozenDict)
        for key in value:
            value_, value_converted = _convert_frozendicts_to_maps(value[key])
            key, key_converted = _convert_frozendicts_to_maps(key)
            map_[key] = value_
            if key_converted or value_converted:
                converted = True
        if converted:
            return map_.dict(), True

    return value, False




import io, binascii
from .read import *

__all__ = (
    'decode',
    'decode_yaml',
    'decode_json',
    'decode_cjson',
    'decode_xml',
    'decode_cbor')


def decode(code, format='yaml'):
    if (format == 'yaml') or (format == ''):
        return decode_yaml(code)
    elif format == 'json':
        return decode_json(code)
    elif format == 'cjson':
        return decode_cjson(code)
    elif format == 'xml':
        return decode_xml(code)
    elif format == 'cbor':
        return decode_cbor(code)
    else:
        raise Exception('unsupported format: ' + format)

def decode_yaml(code):
    return read_yaml(io.StringIO(code))

def decode_json(code):
    return read_json(io.StringIO(code))

def decode_cjson(code):
    return read_cjson(io.StringIO(code))

def decode_xml(code):
    return read_xml(io.StringIO(code))

def decode_cbor(code):
    bytes_ = binascii.a2b_base64(code)
    return read_cbor(io.BytesIO(bytes_))




import ruamel.yaml, json, cbor2
from .types import *
from .yaml import represent_uinteger, represent_map
from .json import Encoder as JSONEncoder
from .cjson import convert_to as convert_to_cjson
from .cbor import encoder_default as cbor_encoder_default

__all__ = (
    'write',
    'write_yaml',
    'write_json',
    'write_cjson',
    'write_xml',
    'write_cbor')


def write(value, writer, format='yaml', indent='', strict=False):
    if (format == 'yaml') or (format == ''):
        write_yaml(value, writer, indent, strict)
    elif format == 'json':
        write_json(value, writer, indent)
    elif format == 'cjson':
        write_cjson(value, writer, indent)
    elif format == 'xml':
        write_xml(value, writer, indent)
    elif format == 'cbor':
        write_cbor(value, writer)
    else:
        raise Exception('unsupported format: ' + format)

def write_yaml(value, writer, indent='', strict=False):
    yaml=ruamel.yaml.YAML(typ='safe')
    yaml.indent = len(indent)
    yaml.representer.add_representer(UInteger, represent_uinteger)
    yaml.representer.add_representer(Map, represent_map)
    yaml.dump(value, writer)

def write_json(value, writer, indent=''):
    if indent == '':
        indent = None
    json.dump(value, writer, ensure_ascii=False, indent=indent, cls=JSONEncoder)
    writer.write('\n')

def write_cjson(value, writer, indent=''):
    value = convert_to_cjson(value)
    write_json(value, writer, indent)

def write_xml(value, writer, indent=''):
    # TODO
    raise NotImplementedError()

def write_cbor(value, writer):
    encoder = cbor2.CBOREncoder(writer, default=cbor_encoder_default)
    encoder.encode(value)




import ruamel.yaml
from .types import *

__all__ = (
    'SafeConstructor',
    'represent_uinteger')


class SafeConstructor(ruamel.yaml.constructor.SafeConstructor):
    '''
    A SafeConstructor that uses the Map class, allowing it to support YAML maps with complex keys.
    '''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.yaml_base_dict_type = Map


class Constructor(ruamel.yaml.constructor.Constructor):
    '''
    A Constructor that uses the Map class, allowing it to support YAML maps with complex keys.
    '''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.yaml_base_dict_type = Map


class RoundTripConstructor(ruamel.yaml.constructor.RoundTripConstructor):
    '''
    A RoundTripConstructor that uses the Map class, allowing it to support YAML maps with complex keys.
    '''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.yaml_base_dict_type = Map


def represent_uinteger(representer, data):
    return representer.represent_int(data)

def represent_map(representer, data):
    return representer.represent_dict(data)




import collections, binascii
from .types import *

__all__ = (
    'CJSON_INTEGER_CODE',
    'CJSON_UINTEGER_CODE',
    'CJSON_BYTES_CODE',
    'CJSON_MAP_CODE',
    'convert_to',
    'convert_from')

CJSON_INTEGER_CODE = '$ard.integer'
CJSON_UINTEGER_CODE = '$ard.uinteger'
CJSON_BYTES_CODE = '$ard.bytes'
CJSON_MAP_CODE = '$ard.map'


def convert_to(value):
    value, _ = _convert_to(value)
    return value

def _convert_to(value):
    if isinstance(value, list):
        converted = False
        list_ = []
        for value_ in value:
            value_, value_converted = _convert_to(value_)
            if value_converted:
                converted = True
            list_.append(value_)
        if converted:
            return list_, True

    elif isinstance(value, tuple):
        converted = False
        list_ = []
        for value_ in value:
            value_, value_converted = _convert_to(value_)
            if value_converted:
                converted = True
            list_.append(value_)
        if converted:
            return tuple(list_), True

    elif isinstance(value, collections.abc.Mapping):
        if len(value) == 1:
            # Check if we need escaping
            try:
                value_, _ = _convert_to(value[CJSON_INTEGER_CODE])
                return {'$'+CJSON_INTEGER_CODE: value_}, True
            except KeyError:
                pass
            try:
                value_, _ = _convert_to(value[CJSON_UINTEGER_CODE])
                return {'$'+CJSON_UINTEGER_CODE: value_}, True
            except KeyError:
                pass
            try:
                value_, _ = _convert_to(value[CJSON_BYTES_CODE])
                return {'$'+CJSON_BYTES_CODE: value_}, True
            except KeyError:
                pass
            try:
                value_, _ = _convert_to(value[CJSON_MAP_CODE])
                return {'$'+CJSON_MAP_CODE: value_}, True
            except KeyError:
                pass

        converted = False
        use_list = False
        dict_ = {}
        list_ = []
        for key in value:
            value_, value_converted = _convert_to(value[key])
            if value_converted:
                converted = True
            if isinstance(key, str):
                list_.append({'key': key, 'value': value_})
                if not use_list:
                    # We can stop building the dict_ if we switched to the list
                    dict_[key] = value_
            else:
                key, _ = _convert_to(key)
                list_.append({'key': key, 'value': value_})
                use_list = True
                converted = True
        if converted:
            if use_list:
                return {CJSON_MAP_CODE: list_}, True
            else:
                return dict_, True
        elif isinstance(value, Map):
            # Maps are not JSON-serializable, so we need the conversion
            return dict_, True

    elif isinstance(value, UInteger): # must be before checking for 'int'
        return {CJSON_UINTEGER_CODE: str(value)}, True

    elif isinstance(value, int):
        return {CJSON_INTEGER_CODE: str(value)}, True

    elif isinstance(value, bytes):
        return {CJSON_BYTES_CODE: binascii.b2a_base64(value).decode()}, True

    return value, False

def convert_from(value):
    value, _ = _convert_from(value)
    return value

def _convert_from(value):
    if isinstance(value, list):
        converted = False
        list_ = []
        for value_ in value:
            value_, value_converted = _convert_from(value_)
            if value_converted:
                converted = True
            list_.append(value_)
        if converted:
            return list_, True

    elif isinstance(value, tuple):
        converted = False
        list_ = []
        for value_ in value:
            value_, value_converted = _convert_from(value_)
            if value_converted:
                converted = True
            list_.append(value_)
        if converted:
            return tuple(list_), True

    elif isinstance(value, collections.abc.Mapping):
        if len(value) == 1:
            # Check for codes
            try:
                return int(value[CJSON_INTEGER_CODE]), True
            except KeyError:
                try:
                    return UInteger(value[CJSON_UINTEGER_CODE]), True
                except KeyError:
                    try:
                        return binascii.a2b_base64(value[CJSON_BYTES_CODE]), True
                    except KeyError:
                        try:
                            map_ = value[CJSON_MAP_CODE]
                            map__ = Map()
                            for entry in map_:
                                key, _ = _convert_from(entry['key'])
                                value_, _ = _convert_from(entry['value'])
                                map__[key] = value_
                            return map__.dict(), True
                        except KeyError:
                            # Handle escape code:
                            # $$ -> $
                            for key, value_ in value.items():
                                if key[:2] == '$$':
                                    key = key[1:]
                                    value_, _ = _convert_from(value_)
                                    return {key: value_}, True

        converted = False
        map_ = Map()
        for key, value_ in value.items():
            key, key_converted = _convert_from(key)
            value_, value_converted = _convert_from(value_)
            if key_converted or value_converted:
                converted = True
            map_[key] = value_
        if converted:
            return map_.dict(), True

    return value, False




import ruamel.yaml, json, cbor2
from .yaml import SafeConstructor as YAMLSafeConstructor
from .json import Encoder as JSONEncoder
from .cjson import convert_from as convert_from_cjson
from .cbor import convert_frozendicts_to_maps

__all__ = (
    'read',
    'read_yaml',
    'read_json',
    'read_cjson',
    'read_xml',
    'read_cbor')


def read(reader, format='yaml'):
    if (format == 'yaml') or (format == ''):
        return read_yaml(reader)
    elif format == 'json':
        return read_json(reader)
    elif format == 'cjson':
        return read_cjson(reader)
    elif format == 'xml':
        return read_xml(reader)
    elif format == 'cbor':
        return read_cbor(reader)
    else:
        raise Exception('unsupported format: ' + format)

def read_yaml(reader):
    yaml=ruamel.yaml.YAML(typ='safe')
    yaml.Constructor = YAMLSafeConstructor
    return yaml.load(reader)

def read_json(reader):
    return json.load(reader)

def read_cjson(reader):
    cjson = read_json(reader)
    return convert_from_cjson(cjson)

def read_xml(reader):
    # TODO
    raise NotImplementedError()

def read_cbor(reader):
    decoder = cbor2.CBORDecoder(reader)
    return convert_frozendicts_to_maps(decoder.decode())




import io, binascii
from .write import *

__all__ = (
    'encode',
    'encode_yaml',
    'encode_json',
    'encode_cjson',
    'encode_xml',
    'encode_cbor')


def encode(value, format='yaml', indent='', strict=False):
    if (format == 'yaml') or (format == ''):
        return encode_yaml(value, indent, strict)
    elif format == 'json':
        return encode_json(value, indent)
    elif format == 'cjson':
        return encode_cjson(value, indent)
    elif format == 'xml':
        return encode_xml(value, indent)
    elif format == 'cbor':
        return encode_cbor(value)
    else:
        raise Exception('unsupported format: ' + format)

def encode_yaml(value, indent='', strict=False):
    buffer = io.StringIO()
    write_yaml(value, buffer, indent, strict)
    return buffer.getvalue()

def encode_json(value, indent=''):
    buffer = io.StringIO()
    write_json(value, buffer, indent)
    return buffer.getvalue()

def encode_cjson(value, indent=''):
    buffer = io.StringIO()
    write_cjson(value, buffer, indent)
    return buffer.getvalue()

def encode_xml(value, indent=''):
    buffer = io.StringIO()
    write_xml(value, buffer, indent)
    return buffer.getvalue()

def encode_cbor(value):
    buffer = io.BytesIO()
    write_cbor(value, buffer)
    return binascii.b2a_base64(buffer.getvalue()).decode()




import json
from .types import *

__all__ = ('Encoder',)


class Encoder(json.JSONEncoder):
    '''
    A JSONEncoder that supports the Map class.
    '''
    def default(self, obj):
        if isinstance(obj, Map):
            return obj.entries
        return json.JSONEncoder.default(self, obj)




import collections.abc

__all__ = (
    'UInteger',
    'Map')


class UInteger(int):
    '''
    An int that will be marked as unsigned where necessary.
    '''


class Map(collections.abc.MutableMapping):
    '''
    A dict-like object that supports arbitrary, even unhashable keys.
    The cost is worst-case performance: implemented as a list rather than a hashtable.
    Iteration retains insertion order.
    Not thread-safe.
    '''

    def __init__(self, entries=None, **kwargs):
        self.entries = []
        self.update(entries, **kwargs)

    def dict(self, strict=False):
        '''
        Attempt to convert to a dict.
        If cannot be converted (due to an unhashable key) will return self
        if strict is False, otherwise will raise a TypeError
        '''
        dict_ = {}
        for key, value in self.entries:
            try:
                dict_[key] = value
            except TypeError:
                # Unhashable key, so cannot be converted to a dict
                if strict:
                    raise
                else:
                    return self
        return dict_

    # JSON serializable

    # Mimic the "dict" contract
    # See: https://docs.python.org/3/library/stdtypes.html#dict

    def keys(self):
        return _MapKeys(self)

    def values(self):
        return _MapValues(self)

    def items(self):
        return _MapItems(self)

    def get(self, key, default=None):
        for key_, value in self.entries:
            if key_ == key:
                return value
        return default

    def pop(self, key, **kwargs):
        for index, (key_, value) in enumerate(self.entries):
            if key_ == key:
                del self.entries[index]
                return value
        try:
            return kwargs['default']
        except KeyError:
            raise KeyError(key)

    def popitem(self):
        return self.entries.pop()

    def setdefault(self, key, default=None):
        for key_, value in self.entries:
            if key_ == key:
                return value
        self.entries.append((key, default))
        return default

    def update(self, other=None, **kwargs):
        if other is not None:
            try:
                # List of tuples
                for key, value in other:
                    self.__setitem__(key, value)
            except (ValueError, TypeError):
                # Dict-like object
                for key in other:
                    self.__setitem__(key, other[key])

        for key, value in kwargs.items():
            self.__setitem__(key, value)
        
        return None

    def clear(self):
        self.entries = []

    def copy(self):
        copy = Map()
        copy.entries = list(self.entries)
        return copy

    # See: https://docs.python.org/3/reference/datamodel.html

    def __hash__(self):
        return hash(tuple(self.entries))

    def __len__(self):
        return len(self.entries)

    def __contains__(self, key):
        for key_, _ in self.entries:
            if key_ == key:
                return True
        return False

    def __getitem__(self, key):
        for key_, value in self.entries:
            if key_ == key:
                return value
        raise KeyError(key)

    def __setitem__(self, key, value):
        for index, entry in enumerate(self.entries):
            if entry[0] == key:
                entry[1] = value
                return
        self.entries.append((key, value))

    def __delitem__(self, key):
        for index, (key_, _) in enumerate(self.entries):
            if key_ == key:
                del self.entries[index]
                return

    def __iter__(self):
        for key, _ in self.entries:
            yield key

    def __reversed__(self):
        for key, _ in reversed(self.entries):
            yield key

    def __or__(self, other): # self | other
        copy = self.copy()
        copy.update(other)
        return copy

    def __ror__(self, other): # other | self
        other = Map(other)
        other.update(self)
        return other

    def __ior__(self, other): # self |= other
        self.update(other)
        return self

    def __eq__(self, other):
        '''
        Equality does not take insertion order into consideration.
        '''
        if not isinstance(other, collections.abc.Mapping):
            return False
        if len(self.entries) != len(other):
            return False
        for key, value in self.entries:
            try:
                if value != other[key]:
                    return False
            except (KeyError, TypeError):
                return False
        return True

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '{' + ', '.join((repr(k) + ': ' + repr(v) for k, v in self.entries)) + '}'

collections.abc.MutableMapping.register(Map)

# Views
# See: https://docs.python.org/3/library/stdtypes.html#dict-views

class _MapKeys(collections.abc.KeysView):
    def __contains__(self, key):
        for key_, _ in self._mapping.entries:
            if key_ == key:
                return True
        return False

    def __iter__(self):
        for key, _ in self._mapping.entries:
            yield key

    def __reversed__(self):
        for key, _ in reversed(self._mapping.entries):
            yield key

collections.abc.KeysView.register(_MapKeys)

class _MapValues(collections.abc.ValuesView):
    def __contains__(self, value):
        for _, value_ in self._mapping.entries:
            if value_ == value:
                return True
        return False

    def __iter__(self):
        for _, value in self._mapping.entries:
            yield value

    def __reversed__(self):
        for _, value in reversed(self._mapping.entries):
            yield value

collections.abc.ValuesView.register(_MapValues)

class _MapItems(collections.abc.ItemsView):
    def __contains__(self, entry):
        return entry in self._mapping.entries

    def __iter__(self):
        yield from self._mapping.entries

    def __reversed__(self):
        yield from reversed(self._mapping.entries)

collections.abc.ItemsView.register(_MapItems)



#!/usr/bin/env python3

import os.path, setuptools

with open(os.path.join(os.path.dirname(__file__), 'VERSION')) as f:
    version = f.read()

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as f:
    readme = f.read()

setuptools.setup(
    name='ard',
    version=version,
    description='Agnostic Raw Data for Python',
    long_description=readme,
    long_description_content_type='text/markdown',
    license='Apache License 2.0',
    author='Tal Liron',
    author_email='tal.liron@gmail.com',
    url='https://github.com/tliron/python-ard',
    download_url='https://github.com/tliron/python-ard/releases',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License'],

    packages=['ard'],
    install_requires=['ruamel.yaml', 'cbor2'])
